#include <stdlib.h>
#include <stdio.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include "aswm.h"
#include "event_handlers/create.h"
#include "event_handlers/configure.h"
#include "event_handlers/map.h"
#include "event_handlers/unmap.h"
#include "event_handlers/reparent.h"
#include "event_handlers/destroy.h"
#include "event_handlers/property.h"
#include "event_handlers/message.h"
#include "log/error.h"

Aswm _aswm;


void aswm_open() {
	_aswm.display = XOpenDisplay(NULL);
	_aswm.desktops = malloc(0);
	_aswm.desktops_count = 0;

	XSetErrorHandler(aswm_error_handler);

	Window x_root_window = DefaultRootWindow(_aswm.display);
	XWindowAttributes x_root_attributes;
	XGetWindowAttributes(_aswm.display, x_root_window, &x_root_attributes);

	_aswm.root_window = XCreateSimpleWindow(
			_aswm.display, x_root_window,
			0, 0,
			x_root_attributes.width,
			x_root_attributes.height,
			0, 0, 0);

	// User windows are created from the X root window, while ASWM specific
	// windows are created on the ASWM root window
	XSelectInput(_aswm.display, _aswm.root_window, FocusChangeMask | EnterWindowMask | LeaveWindowMask | ButtonPressMask | ButtonReleaseMask);

	XSelectInput(_aswm.display, x_root_window, SubstructureRedirectMask | SubstructureNotifyMask | FocusChangeMask | EnterWindowMask | LeaveWindowMask | ButtonPressMask | ButtonReleaseMask);

	XMapWindow(_aswm.display, _aswm.root_window);
	XSync(_aswm.display, 0);

	printf("PointerRoot %lu\n", PointerRoot);
	printf("Useful constants:\n");
	printf("\tNone: %lu\n", None);
	printf("\tTrue / False: %i / %i\n", True, False);
	printf("X Root window %lu\n", x_root_window);
	printf("ASWM Root window %lu\n", _aswm.root_window);

	atexit(aswm_close);
}

void aswm_create_workspace() {
	_aswm.desktops = realloc(_aswm.desktops, (_aswm.desktops_count+1) * sizeof(Desktop));

	XWindowAttributes root_attributes;
	XGetWindowAttributes(_aswm.display, _aswm.root_window, &root_attributes);

	_aswm.desktops[_aswm.desktops_count].root_window = XCreateSimpleWindow(
			_aswm.display, _aswm.root_window,
			0, 0,
			root_attributes.width, root_attributes.height,
			0, 0, 0);
	_aswm.desktops[_aswm.desktops_count].tile_root = XCreateSimpleWindow(
			_aswm.display, _aswm.desktops[_aswm.desktops_count].root_window,
			0, 0,
			root_attributes.width, root_attributes.height,
			0, 0, 0);
	_aswm.desktops[_aswm.desktops_count].stack_root = XCreateSimpleWindow(
			_aswm.display, _aswm.desktops[_aswm.desktops_count].root_window,
			0, 0,
			root_attributes.width, root_attributes.height,
			0, 0, 0);

	_aswm.current_desktop = _aswm.desktops[_aswm.desktops_count];
	_aswm.desktops_count++;
	XSelectInput(_aswm.display, _aswm.current_desktop.tile_root, SubstructureNotifyMask);
	XSelectInput(_aswm.display, _aswm.current_desktop.stack_root, SubstructureNotifyMask);
	XMapWindow(_aswm.display, _aswm.current_desktop.root_window);
	XMapWindow(_aswm.display, _aswm.current_desktop.tile_root);
	printf("Create workspace %lu\n", _aswm.current_desktop.root_window);
}

void aswm_event_loop() {
	printf("MotionNofity %i\n", MotionNotify);
	printf("ButtonPress %i\n", ButtonPress);
	printf("ButtonRelease %i\n", ButtonRelease);
	printf("ColormapNotify %i\n", ColormapNotify);
	printf("EnterNotify %i\n", EnterNotify);
	printf("LeaveNotify %i\n", LeaveNotify);
	printf("Expose %i\n", Expose);
	printf("GraphicsExpose %i\n", GraphicsExpose);
	printf("NoExpose %i\n", NoExpose);
	printf("FocusIn %i\n", FocusIn);
	printf("FocusOut %i\n", FocusOut);
	printf("KeymapNotify %i\n", KeymapNotify);
	printf("KeyPress %i\n", KeyPress);
	printf("KeyRelease %i\n", KeyRelease);
	printf("MotionNofity %i\n", MotionNotify);
	printf("PropertyNotify %i\n", PropertyNotify);
	printf("ResizeRequest %i\n", ResizeRequest);
	printf("CirculateNotify %i\n", CirculateNotify);
	printf("ConfigureNotify %i\n", ConfigureNotify);
	printf("DestroyNotify %i\n", DestroyNotify);
	printf("GravityNotify %i\n", GravityNotify);
	printf("MapNotify %i\n", MapNotify);
	printf("ReparentNotify %i\n", ReparentNotify);
	printf("UnmapNotify %i\n", UnmapNotify);
	printf("CirculateRequest %i\n", CirculateRequest);
	printf("ConfigureRequest %i\n", ConfigureRequest);
	printf("MapRequest %i\n", MapRequest);
	printf("VisibilityNotify %i\n", VisibilityNotify);

	for (;;) {
		XEvent e;
		XNextEvent(_aswm.display, &e);

		switch(e.type) {
			case CreateNotify:
				// Windows with override_redirect set to True can be safely
				// ignored, according to the X server specification.
				if(e.xcreatewindow.override_redirect == False)
					OnCreateNotify(&_aswm, &e.xcreatewindow);
				else
					printf("Ignored create notify %lu\n", e.xcreatewindow.window);
				break;
			case ConfigureRequest:
				OnConfigureRequest(&e.xconfigurerequest);
				break;
			case MapRequest:
				OnMapRequest(&_aswm, &e.xmaprequest);
				break;
			case UnmapNotify:
				OnUnmapNotify(&_aswm, &e.xunmap);
				break;
			case DestroyNotify:
				OnDestroyNotify(&e.xdestroywindow);
				break;
			case ReparentNotify:
				OnReparentNotify(&e.xreparent);
				break;
			case FocusIn:
				printf("FocusIn %lu\n", (&e.xfocus)->window);
				break;
			case FocusOut:
				printf("FocusOut %lu\n", e.xfocus.window);
				break;
			case EnterNotify:
				int revert_to;
				Window window;
				XGetInputFocus(e.xcrossing.display, &window, &revert_to);
				printf("Enter %lu\n", e.xcrossing.window);
				printf("Input focus %lu\n", window);
				break;
			case LeaveNotify:
				printf("Leave %lu\n", e.xcrossing.window);
				break;
			case ButtonPress:
				printf("Press %lu\n", e.xbutton.window);
				break;
			case ButtonRelease:
				printf("Release %lu\n", e.xbutton.window);
				break;
			case ClientMessage:
				OnClientMessage(&_aswm, &e.xclient);
				break;
			case PropertyNotify:
				OnPropertyNotify(&_aswm, &e.xproperty);
				break;
			default:
				printf("Unknown event: %i\n", e.type);
				break;
					/* LOG(WARNING) << "Ignored event"; */
		}
	}
}

void aswm_close() {
	free(_aswm.desktops);
	XCloseDisplay(_aswm.display);
}

Window aswm_currently_active() {
	Window active;
	int status;
	XGetInputFocus(_aswm.display, &active, &status);
	return active;
}

void aswm_activate_window(Window window) {
	/* XEvent event; */
	/* event.xclient.display = _aswm.display; */
	/* event.xclient.window = window; */
	/* event.xclient.message_type = XInternAtom(_aswm.display, "_NET_ACTIVE_WINDOW", True); */
	/* event.xclient.format = 32; */
	/* event.xclient.data.l[0] = 1; // Source indication */
	/* event.xclient.data.l[1] = 1; // Timestamp // Currently not set */
	/* event.xclient.data.l[2] = 0; // Requestor's currently active window: TODO */
	
	printf("Activate window %lu (currently active: %lu)\n", window, aswm_currently_active());
	/* XSendEvent(_aswm.display, _aswm.root_window, True, 0 [> N. A. <], &event); */
	XSetInputFocus(_aswm.display, window, RevertToParent, CurrentTime);
}

int main(int argc, char** argv) {
	aswm_open();
	aswm_event_loop();
	aswm_close();
}
